/* $Id$

This file is part of gldbg
Copyright © 2009-2011  Mikko Rasa, Mikkosoft Productions
Distributed under the GPL
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <dlfcn.h>
#include <fcntl.h>
#include <signal.h>
#include "arraysize.h"
#include "breakpoint.h"
#include "functions.h"
#include "opengl.h"
#include "packet.h"

#define UNUSED __attribute__((unused))

static unsigned char breakpoints[N_FUNCS] = { };
static char break_any = 0;
static char hold = 0;
static char no_break = 0;

static const char *get_lib_names(void)
{
	const char *env = getenv("GLWRAP_LIBS");
	if(env)
		return env;
	return "libGL.so.1";
}

void *glsym(const char *name)
{
	static void **gl_libs = NULL;
	unsigned i;

	if(!gl_libs)
	{
		char *lib_names = strdup(get_lib_names());
		unsigned n_libs = 1;
		unsigned j;

		for(i=0; lib_names[i]; ++i)
			if(lib_names[i]==':')
				++n_libs;

		gl_libs = (void **)malloc((n_libs+1)*sizeof(void *));
		i = 0;
		n_libs = 0;
		for(j=0;; ++j)
		{
			if(lib_names[j]==':' || lib_names[j]==0)
			{
				int at_end = (lib_names[j]==0);
				lib_names[j] = 0;

				gl_libs[n_libs] = dlopen(lib_names+i, RTLD_NOW);
				if(!gl_libs[n_libs])
				{
					fprintf(stderr, "Could not open %s: %s\n", lib_names+i, dlerror());
					abort();
				}

				i = j+1;
				++n_libs;

				if(at_end)
					break;
			}
		}

		gl_libs[n_libs] = 0;
		free(lib_names);
	}

	for(i=0; gl_libs[i]; ++i)
	{
		void *sym = dlsym(gl_libs[i], name);
		if(sym)
			return sym;
	}

	return NULL;
}

int get_out_fd(void)
{
	static int fd = -1;

	if(fd<0)
	{
		const char *var = getenv("GLWRAP_FD");
		if(var)
			fd = strtol(var, NULL, 0);
		else
		{
			var = getenv("GLWRAP_FILE");
			if(var)
			{
				fd = open(var, O_WRONLY|O_CREAT|O_TRUNC, 0644);
				if(fd==-1)
				{
					fprintf(stderr, "Couldn't open dumpfile %s for output: %s", var, strerror(errno));
					abort();
				}
			}
			else
				fd = 2;
		}
	}

	return fd;
}

int get_in_fd(void)
{
	static int fd = -1;

	if(fd<0)
	{
		const char *var = getenv("GLWRAP_CTRL_FD");
		if(var)
			fd = strtol(var, NULL, 0);
		else
			fd = 0;
	}

	return fd;
}

static void receive_gldBreak(GlPacket *pkt)
{
	unsigned short func;
	unsigned char flags_set;
	unsigned char flags_clr;

	packet_read_short(pkt, (short *)&func);
	packet_read_char(pkt, (char *)&flags_set);
	packet_read_char(pkt, (char *)&flags_clr);

	if(func)
	{
		breakpoints[func] &= ~flags_clr;
		breakpoints[func] |= flags_set;
	}
	else
		break_any = flags_set;
}

static void receive_gldHold(GlPacket *pkt UNUSED)
{
	hold = 1;
}

static void receive_gldQueryViewport(GlPacket *pkt UNUSED)
{
	int viewport[4];

	no_break = 1;
	glGetIntegerv(GL_VIEWPORT, viewport);
	no_break = 0;
}

static void receive_gldReadPixels(GlPacket *pkt)
{
	GLint x, y;
	GLsizei width, height;
	GLenum format, type;
	char *data;

	packet_read_int(pkt, &x);
	packet_read_int(pkt, &y);
	packet_read_int(pkt, &width);
	packet_read_int(pkt, &height);
	packet_read_int(pkt, (int *)&format);
	packet_read_int(pkt, (int *)&type);

	data = (char *)malloc(width*height*typesize(type)*formatsize(format));
	no_break = 1;
	glReadPixels(x, y, width, height, format, type, data);
	no_break = 0;
	free(data);
}

static void receive(void)
{
	GlPacket *pkt;

	while((pkt = packet_receive(get_in_fd())))
	{
		unsigned short func;

		packet_read_short(pkt, (short *)&func);
		switch(func)
		{
		case FUNC_GLDBREAK: receive_gldBreak(pkt); break;
		case FUNC_GLDHOLD: receive_gldHold(pkt); break;
		case FUNC_GLDQUERYVIEWPORT: receive_gldQueryViewport(pkt); break;
		case FUNC_GLDREADPIXELS: receive_gldReadPixels(pkt); break;
		default:;
		}
	}
}

void tracepoint(unsigned short func, int flag)
{
	int breakpoint;

	if(no_break)
		return;

	receive();

	breakpoint = (breakpoints[func]|break_any)&flag;
	if(breakpoint || hold)
	{
		GlPacket *pkt;

		if(breakpoint)
		{
			pkt = packet_begin(FUNC_GLDBREAK);
			packet_write_short(pkt, func);
			packet_write_char(pkt, flag);
			packet_send(pkt, get_out_fd());
		}

		break_any = 0;

		while(1)
		{
			hold = 0;
			raise(SIGTRAP);
			receive();
			if(!hold)
				break;
		}
	}
}
